{***********************************************************}
{*                    Unit StrmView                        *}
{*       (c) 1997 by Dmitry Beloshistov [-=BDS=-]          *}
{***********************************************************}
unit StrmView;
{$O+,F+,X+,I-,S-}
interface
uses Objects,Drivers,Memory,Views,Strings;

type
  PStreamIndicator=^TStreamIndicator;
  TStreamIndicator=object(TView)
                    Cur,Max:Longint;
                    Constructor Init(var Bounds:TRect; AMax:longint);
                    Procedure Draw; virtual;
                    Procedure SetCurValue(ACur:Longint);
                    Procedure SetMaxValue(AMax:Longint);
                   end;
                   
  PStreamViewer = ^TStreamViewer;
  TStreamViewer = object(TView)
    VScrollBar: PScrollBar;
    NumCols: Integer;
    TopItem: longint;
    Focused: longint;
    Range: longint;
    RecSize:Word;
    NumRecs:Longint;
    Status:integer;
    Modify:boolean;
    S:TBufStream;
    constructor Init(var Bounds:TRect; ANumCols,ARecSize:Word;
                     AVScrollBar:PScrollBar; FileName:String;
                     AModify:boolean; BufSize:Word);
    Destructor Done; virtual;
    procedure ChangeBounds(var Bounds: TRect); virtual;
    procedure Draw; virtual;
    procedure FocusItem(Item: LongInt); virtual;
    function GetPalette: PPalette; virtual;
    function GetText(Item:Longint; MaxLen: Integer): String; virtual;
    function IsSelected(Item:Longint): Boolean; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure SelectItem(Item: Longint); virtual;
    procedure SetRange(ARange: longint);
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure PutRec(RecNumber:Longint; var Rec);
    procedure GetRec(RecNumber:longint; var Rec);
    Procedure AppendRec(var Rec);
  private
    procedure FocusItemNum(Item: longint); virtual;
  end;

Implementation
{------------------------ TStreamIndicator -------------------------}
Constructor TStreamIndicator.Init(var Bounds:TRect; AMax:longint);
begin
 Inherited Init(Bounds);
 Cur:=0; Max:=AMax;
end;
Procedure TStreamIndicator.SetCurValue(ACur:Longint);
begin
 Cur:=ACur;
 Draw;
end;
Procedure TStreamIndicator.SetMaxValue(AMax:Longint);
begin
 Max:=AMax;
 Draw;
end;
Procedure TStreamIndicator.Draw;

   Function LZ(S:String):String;
   begin
    while Pos(' ',s)<>0 do S[Pos(' ',S)]:='0';
    LZ:=S;
   end;

var S1,S2:String;
    B:TDrawBuffer;
    Color:Word;
begin
 Color:=GetColor(2);
 MoveChar(B,' ',Color,Size.X);
 Str((Cur+1):6,S1); Str(Max:6,S2); S1:='['+lz(S1)+'/'+lz(S2)+']';
 MoveStr(B,S1,Color);
 WriteLine(0,0,Size.X,1,B);
end;
{--------------------------- TStreamViewer -------------------------}
constructor TStreamViewer.Init(var Bounds:TRect; ANumCols,ARecSize:Word;
                             AVScrollBar:PScrollBar; FileName:String;
                             AModify:Boolean; BufSize:Word);
var ArStep, PgStep: Integer;
begin
  TView.Init(Bounds);
  Options := Options or (ofFirstClick + ofSelectable);
  EventMask := EventMask or evBroadcast;
  Range := 0;
  NumCols := ANumCols;
  Focused := 0;
  if AVScrollBar <> nil then
  begin
    if NumCols = 1 then
    begin
      PgStep := Size.Y -1;
      ArStep := 1;
    end else
    begin
      PgStep := Size.Y * NumCols;
      ArStep := Size.Y;
    end;
    AVScrollBar^.SetStep(PgStep, ArStep);
  end;
  VScrollBar := AVScrollBar;
  Modify:=AModify;
  if not Modify then S.Init(FileName,stOpenRead,BufSize)
  else S.Init(FileName,stOpen,BufSize);
  if S.Status<>stOk then Fail;
  RecSize:=ARecSize;
  NumRecs:=S.GetSize div RecSize;
  Status:=0;
  SetRange(NumRecs);
end;

Destructor TStreamViewer.Done;
begin
 S.Done;
 Inherited Done;
end;

procedure TStreamViewer.ChangeBounds(var Bounds: TRect);
begin
  TView.ChangeBounds(Bounds);
  if VScrollBar <> nil then VScrollBar^.SetStep(Size.Y, VScrollBar^.ArStep);
end;

procedure TStreamViewer.Draw;
var
  I, J, Item: Integer;
  NormalColor, SelectedColor, FocusedColor, Color: Word;
  ColWidth, CurCol, Indent: Integer;
  B: TDrawBuffer;
  Text: String;
  SCOff: Byte;
begin
  if State and (sfSelected + sfActive) = (sfSelected + sfActive) then
  begin
    NormalColor := GetColor(1);
    FocusedColor := GetColor(3);
    SelectedColor := GetColor(4);
  end else
  begin
    NormalColor := GetColor(2);
    SelectedColor := GetColor(4);
  end;
  Indent := 0;
  ColWidth := Size.X div NumCols + 1;
  for I := 0 to Size.Y - 1 do
  begin
    for J := 0 to NumCols-1 do
    begin
      Item := J*Size.Y + I + TopItem;
      CurCol := J*ColWidth;
      if (State and (sfSelected + sfActive) = (sfSelected + sfActive)) and
        (Focused = Item) and (Range > 0) then
      begin
        Color := FocusedColor;
        SetCursor(CurCol+1,I);
        SCOff := 0;
      end
      else if (Item < Range) and IsSelected(Item) then
      begin
        Color := SelectedColor;
        SCOff := 2;
      end
      else
      begin
        Color := NormalColor;
        SCOff := 4;
      end;
      MoveChar(B[CurCol], ' ', Color, ColWidth);
      if Item < Range then
      begin
        Text := GetText(Item, ColWidth + Indent);
        Text := Copy(Text,Indent,ColWidth);
        MoveStr(B[CurCol+1], Text, Color);
        if ShowMarkers then
        begin
          WordRec(B[CurCol]).Lo := Byte(SpecialChars[SCOff]);
          WordRec(B[CurCol+ColWidth-2]).Lo := Byte(SpecialChars[SCOff+1]);
        end;
      end;
      MoveChar(B[CurCol+ColWidth-1], #179, GetColor(5), 1);
    end;
    WriteLine(0, I, Size.X, 1, B);
  end;
end;

procedure TStreamViewer.FocusItem(Item: longint);
begin
  Focused := Item;
  if VScrollBar <> nil then VScrollBar^.SetValue(Item);
  if Item < TopItem then
    if NumCols = 1 then TopItem := Item
    else TopItem := Item - Item mod Size.Y
  else if Item >= TopItem + (Size.Y*NumCols) then
    if NumCols = 1 then TopItem := Item - Size.Y + 1
    else TopItem := Item - Item mod Size.Y - (Size.Y*(NumCols - 1));
end;

procedure TStreamViewer.FocusItemNum(Item: longint);
begin
  if Item < 0 then Item := 0
  else if (Item >= Range) and (Range > 0) then Item := Range-1;
  if Range <> 0 then FocusItem(Item);
end;

function TStreamViewer.GetPalette: PPalette;
const
  P: String[Length(CListViewer)] = CListViewer;
begin
  GetPalette := @P;
end;

function TStreamViewer.GetText(Item: LongInt; MaxLen: Integer): String;
var Buf:array[0..255] of char;
begin
  if MaxLen>RecSize then MaxLen:=RecSize;
  if MaxLen>255 then MaxLen:=255;
  S.Seek(Item*RecSize);
  S.Read(Buf,MaxLen); Buf[MaxLen]:=#0;
  GetText:=copy(StrPas(Buf),1,MaxLen);
end;

function TStreamViewer.IsSelected(Item: LongInt): Boolean;
begin
  IsSelected := Item = Focused;
end;

procedure TStreamViewer.HandleEvent(var Event: TEvent);
const
  MouseAutosToSkip = 4;
var
  Mouse: TPoint;
  ColWidth: Word;
  OldItem, NewItem: Integer;
  Count: Word;
begin
  TView.HandleEvent(Event);
  if Event.What = evMouseDown then
  begin
    ColWidth := Size.X div NumCols + 1;
    OldItem := Focused;
    MakeLocal(Event.Where, Mouse);
    if MouseInView(Event.Where) then
      NewItem := Mouse.Y + (Size.Y * (Mouse.X div ColWidth)) + TopItem
    else NewItem := OldItem;
    Count := 0;
    repeat
      if NewItem <> OldItem then
      begin
        FocusItemNum(NewItem);
        DrawView;
      end;
      OldItem := NewItem;
      MakeLocal(Event.Where, Mouse);
      if MouseInView(Event.Where) then
        NewItem := Mouse.Y + (Size.Y * (Mouse.X div ColWidth)) + TopItem
      else
      begin
        if NumCols = 1 then
        begin
          if Event.What = evMouseAuto then Inc(Count);
          if Count = MouseAutosToSkip then
          begin
            Count := 0;
            if Mouse.Y < 0 then NewItem := Focused-1
            else if Mouse.Y >= Size.Y then NewItem := Focused+1;
          end;
        end
        else
        begin
          if Event.What = evMouseAuto then Inc(Count);
          if Count = MouseAutosToSkip then
          begin
            Count := 0;
            if Mouse.X < 0 then NewItem := Focused-Size.Y
            else if Mouse.X >= Size.X then NewItem := Focused+Size.Y
            else if Mouse.Y < 0 then
              NewItem := Focused - Focused mod Size.Y
            else if Mouse.Y > Size.Y then
              NewItem := Focused - Focused mod Size.Y + Size.Y - 1;
          end
        end;
      end;
    until not MouseEvent(Event, evMouseMove + evMouseAuto);
    FocusItemNum(NewItem);
    DrawView;
    if Event.Double and (Range > Focused) then SelectItem(Focused);
    ClearEvent(Event);
  end
  else if Event.What = evKeyDown then
  begin
    if (Event.CharCode = ' ') and (Focused < Range) then
    begin
      SelectItem(Focused);
      NewItem := Focused;
    end
    else case CtrlToArrow(Event.KeyCode) of
      kbUp: NewItem := Focused - 1;
      kbDown: NewItem := Focused + 1;
      kbRight: if NumCols > 1 then NewItem := Focused + Size.Y else Exit;
      kbLeft: if NumCols > 1 then NewItem := Focused - Size.Y else Exit;
      kbPgDn: NewItem := Focused + Size.Y * NumCols;
      kbPgUp: NewItem := Focused - Size.Y * NumCols;
      kbHome: NewItem := TopItem;
      kbEnd: NewItem := TopItem + (Size.Y * NumCols) - 1;
      kbCtrlPgDn: NewItem := Range - 1;
      kbCtrlPgUp: NewItem := 0;
    else
      Exit;
    end;
    FocusItemNum(NewItem);
    DrawView;
    ClearEvent(Event);
  end else if Event.What = evBroadcast then
    if Options and ofSelectable <> 0 then
      if (Event.Command = cmScrollBarClicked) and (Event.InfoPtr = VScrollBar) then
            Select
      else if (Event.Command = cmScrollBarChanged) then
      begin
        if (VScrollBar = Event.InfoPtr) then
        begin
          FocusItemNum(VScrollBar^.Value);
          DrawView;
        end;
      end;
end;

procedure TStreamViewer.SelectItem(Item: longint);
begin
  Message(Owner, evBroadcast, cmListItemSelected, @Self);
end;

procedure TStreamViewer.SetRange(ARange: longint);
begin
  Range := ARange;
  if VScrollBar <> nil then
  begin
    if Focused > ARange then Focused := 0;
    VScrollbar^.SetParams(Focused, 0, ARange-1, VScrollBar^.PgStep,
      VScrollBar^.ArStep);
  end;
end;

procedure TStreamViewer.SetState(AState: Word; Enable: Boolean);

procedure ShowSBar(SBar: PScrollBar);
begin
  if (SBar <> nil) then
    if GetState(sfActive) and GetState(sfVisible) then SBar^.Show
    else SBar^.Hide;
end;

begin
  TView.SetState(AState, Enable);
  if AState and (sfSelected + sfActive + sfVisible) <> 0 then
  begin
    ShowSBar(VScrollBar);
    DrawView;
  end;
end;

Procedure TStreamViewer.GetRec(RecNumber:Longint; var Rec);
begin
 S.Seek(RecNumber*RecSize);
 Status:=S.Status;
 S.Read(Rec,RecSize);
 Status:=S.Status;
end;

Procedure TStreamViewer.PutRec(RecNumber:Longint; var Rec);
begin
 if not Modify then exit;
 S.Seek(RecNumber*RecSize);
 Status:=S.Status;
 S.Write(Rec,RecSize);
 Status:=S.Status;
end;

Procedure TStreamViewer.AppendRec(var Rec);
begin
 if not Modify then exit;
 S.Seek(S.GetSize);
 S.Write(Rec,RecSize);
 Status:=S.Status;
end;

end.
